# -*- coding:utf-8 -*-


"""
"""


import os


__author__ = "Andres FR"
__email__ = "aferro@em.uni-frankfurt.de"


with open(os.path.join(os.path.dirname(__file__),
                       "instructions.txt"), "r") as f:
    INSTRUCTIONS = f.read()
with open(os.path.join(os.path.dirname(__file__),
                       "about.txt"), "r") as f:
    ABOUT = f.read()



# -*- coding:utf-8 -*-


"""
This module contains the ``QGraphicsScene+QGraphicsView`` binomial, typically
used in Qt apps to display and navigate images, together with some specific
functionality to annotate.
"""


import numpy as np
from PySide2 import QtCore, QtWidgets, QtGui
from PIL import Image
#
from .utils import RandomColorGenerator, rgb_arr_to_rgb_pixmap, \
    bool_arr_to_rgba_pixmap, pixmap_to_arr, unique_filename
from .mouse_event_manager import MouseEventManager


# #############################################################################
# ## SCENE (PAINT ETC)
# #############################################################################
class MaskedImageScene(QtWidgets.QGraphicsScene):
    """
    Basic area that allows to display a color image, together with a set of
    binary masks on top of it.
    """
    DEFAULT_MASK_ALPHA = 100  # 255 is opaque. Used if no colors are specified

    def __init__(self, img_arr=None, parent=None):
        """
        :param img_arr: See ``update_image``
        """
        super().__init__(parent)
        self.img_pmi = None
        self.h = None
        self.w = None
        #
        self.mask_pmis = {}  # pmi_ref : (r, g, b, a)
        if img_arr is not None:
            self.update_image(img_arr)

    def update_image(self, img_arr):
        """
        Clears whole scene, and adds the given numpy array as Pixmap.

        :param img_arr: A ``np.uint8(h, w [, ?])`` array.
        """
        pm = rgb_arr_to_rgb_pixmap(img_arr)
        self.clear()
        self.img_pmi = self.addPixmap(pm)
        #
        self.h, self.w = img_arr.shape[:2]
        self.setSceneRect(0, 0, self.w, self.h)

    def num_items(self):
        """
        Number of items in this scene, ordered from foreground to background.
        """
        return len(self.items(QtCore.Qt.DescendingOrder))

    def items(self, ascending=True):
        """
        :returns: This scene's items.
        :param ascending: If true, items are given from background to
          foreground.
        """
        o = QtCore.Qt.AscendingOrder if ascending else QtCore.Qt.AscendingOrder
        return super().items(o)

    def add_mask_pmi(self, mask_arr, rgba=None, item_on_top=None):
        """
        :param mask_arr: A ``np.bool(h, w)`` array.
        :param item_on_top: If given, mask will be added underneath that item.
          Otherwise will be added on top of item stack.
        :returns: The added ``PixmapItem``.
        """
        # sanity check mask
        assert mask_arr.dtype == np.bool, "Mask must be np.bool(h, w)!"
        assert len(mask_arr.shape) == 2, "Mask must be np.bool(h, w)!"
        assert mask_arr.shape == (self.h, self.w), \
            "Mask must have same (h, w) as image!"
        # sanity check color
        if rgba is None:
            r, g, b = next(RandomColorGenerator().generate(form="rgbArray"))
            rgba = (r, g, b, self.DEFAULT_MASK_ALPHA)
        assert all([0 <= c <= 255 for c in rgba]), \
            "RGBA must be in [0, 255] range!"
        # add pixmap: if this fails, the method raises with no side effect
        pm = bool_arr_to_rgba_pixmap(mask_arr, rgba)
        pmi = self.addPixmap(pm)
        self.mask_pmis[pmi] = rgba
        # now we have side FX: if this fails roll back the add Pixmap
        if item_on_top is not None:
            try:
                pmi.stackBefore(item_on_top)
            except Exception:
                # roll back addition
                del self.mask_pmis[pmi]
                self.removeItem(pmi)
                raise RuntimeError(
                    "Invalid other_item? {}".format(item_on_top))
        #
        return pmi

    def remove_mask_pmi(self, pmi):
        """
        :param pmi: The PixmapItem to remove. It has to be a mask added via
          ``add_mask``
        """
        # check that exists in our dict
        rgba = self.mask_pmis[pmi]
        #
        self.removeItem(pmi)
        del self.mask_pmis[pmi]
        return rgba

    def replace_mask_pmi(self, pmi, new_mask_arr, new_rgba=None):
        """
        If we call ``remove_mask`` and then ``add_mask`` fails, we will lose
        the removed mask forever. This method updates the mask in an atomary
        way: either succeeds or does nothing.

        .. warning::
          The input ``pmi`` gets removed from the scene and the reference
          becomes invalid. Use the reference returned by this function instead.
        """
        # the problem is if we successfully remove and then fail to add. A
        # fix is to first add below the to-be-removed, and then remove
        old_rgba = self.mask_pmis[pmi]
        if new_rgba is None:
            new_rgba = old_rgba
        new_pmi = self.add_mask_pmi(new_mask_arr, rgba=new_rgba,
                                    item_on_top=pmi)
        self.remove_mask_pmi(pmi)
        return new_pmi

    def pixmap_item_to_np_mask(self, pmi,
                               pixmap_format=QtGui.QImage.Format_RGBA8888):
        """
        Converts given pixmap to np.bool, assuming all non-zero values
        are True. Returns the bool array.
        """
        # a has shape (h, w, 4) where 4->RGBA
        arr = pixmap_to_arr(pmi.pixmap(), pixmap_format)
        mask = (arr > 0).any(axis=-1)
        return mask

    def save_mask_as_image(self, pmi, outpath,
                           overwrite_existing=False, verbose=True):
        """
        Output: RGB PNG image where false is black (0, 0, 0) and true is white
        (255, 255, 255).
        """
        if not overwrite_existing:
            outpath = unique_filename(outpath)
        m = self.pixmap_item_to_np_mask(pmi)
        #
        img = Image.fromarray(m)
        img.save(outpath)
        if verbose:
            print("saved to", outpath)


# #############################################################################
# ## VIEW (EVENTS ETC)
# #############################################################################
class DisplayView(MouseEventManager, QtWidgets.QGraphicsView):
    """
    In Qt applications, it is usual to wrap a scene with a view. This allows
    to dynamically and easily change the perspective on the scene.
    """
    def __init__(self, scene=None, parent=None, scale_percent=15):
        """
        :param scene: The scene to view. It can be set also afterwards via
          ``self.setScene``.
        """
        QtWidgets.QGraphicsView.__init__(self, scene, parent)
        MouseEventManager.__init__(self, True)
        self.scale_factor_up = 1 + float(scale_percent) / 100
        self.scale_factor_down = 1.0 / self.scale_factor_up
        #
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        # Set Anchors
        self.setTransformationAnchor(QtWidgets.QGraphicsView.NoAnchor)
        self.setResizeAnchor(QtWidgets.QGraphicsView.NoAnchor)
        #
        if scene is not None:
            self.fit_in_scene()

    def fit_in_scene(self):
        """
        Moves perspective so that the whole scene can be seen in view.
        """
        self.fitInView(self.scene().sceneRect(), QtCore.Qt.KeepAspectRatio)

    def zoom(self, pos_x, pos_y, zoom_out=False):
        """
        Source for wheel zoom: https://stackoverflow.com/a/29026916/4511978
        """
        # Save the scene pos
        qpos = QtCore.QPoint(pos_x, pos_y)
        old_pos = self.mapToScene(qpos)
        # Zoom
        scale_factor = (self.scale_factor_down if zoom_out
                        else self.scale_factor_up)
        self.scale(scale_factor, scale_factor)
        # Get the new position
        new_pos = self.mapToScene(qpos)
        # Move scene to old position
        delta = new_pos - old_pos
        self.translate(delta.x(), delta.y())

    def shift_view(self, delta_x, delta_y):
        """
        Move perspective to shift through the scene
        """
        self.horizontalScrollBar().setValue(
            self.horizontalScrollBar().value() - delta_x)
        self.verticalScrollBar().setValue(
            self.verticalScrollBar().value() - delta_y)

    def on_mid_press(self, event):
        """
        Override me
        """
        self.fit_in_scene()

    def on_wheel(self, event, has_ctrl, has_alt, has_shift):
        """
        Override me
        """
        if (has_ctrl, has_alt, has_shift) == (False, False, False):
            zoom_out = event.delta() <= 0
            self.zoom(*event.pos().toTuple(), zoom_out)

    def on_right_press(self, event):
        """
        Override me
        """
        self.setDragMode(self.ScrollHandDrag)

    def on_right_release(self, event):
        """
        Override me
        """
        self.setDragMode(self.NoDrag)

    def on_move(self, event, has_left, has_mid, has_right, this_pos, last_pos):
        """
        Override me
        """
        #
        # shift scene with right button
        if has_right:
            delta_x, delta_y = (this_pos - last_pos).toTuple()
            self.shift_view(delta_x, delta_y)



# -*- coding:utf-8 -*-


"""
Deploy script for the package. It has to be in the repository root directory.
Make sure that the config entries are correct!
"""


import setuptools


# See https://stackoverflow.com/a/33685899/4511978
REQUIREMENTS = [
    "flake8",
    "coverage",
    "snakeviz",
    # "memory_profiler",
    "sphinx",
    "sphinx_rtd_theme",
    "bump2version",
    "setuptools",
    "wheel",
    "twine",
    # "mypy",
    #
    "ExifRead",
    "randomcolor",
    "numpy",
    "scikit-image",
    "PySide2"]


def setup():
    """
    The proper setup function, adapted from the tutorial in
    https://packaging.python.org/tutorials/packaging-projects/
    """
    with open("README.md", "r") as f:
        long_description = f.read()
    #
    setuptools.setup(
        name="secv_guis",
        # the version is automatically handled by "bumpversion"
        version="0.2.1",
        author="aferro",
        author_email="aferro@em.uni-frankfurt.de",
        description="Qt GUIs for Systems Engineering and Computer Vision",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/andres-fr/secv-guis",
        packages=setuptools.find_packages(exclude=["*utest*"]),
        # dict of package_name: [list_of_patterns]
        package_data={"secv_guis": ["*/*.txt"]},
        include_package_data=True,
        install_requires=REQUIREMENTS,
        classifiers=[
            # comprehensive list: https://pypi.org/classifiers/
            "Programming Language :: Python :: 3 :: Only",
            # "Programming Language :: Python",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent"
        ],
    )


if __name__ == "__main__":
    setup()



# -*- coding:utf-8 -*-


"""
This module contains a convenience mixin that provides the following
functionality for mouse tracking:

* Record previous mouse states
* Demultiplex mouse events and preprocess relevant informations

To make a widget responsive to mouse events, simply this class there
and override the desired methods.
"""


from PySide2 import QtCore


class MouseEventManager:
    """
    Extend this class and then instantiate it once as a property in your
    desired widget.

    .. warning::
      The Mixin is compatible with multiple inheritance, but not all
      initializations work. The following does:

      class A(MouseEventManager, QtWidgets.XXX):
        def __init__(self, ...):
          QtWidgets.XXX.__init__(self, ...)
          MouseEventManager.__init__(self, ...)

    In this example, class ``A`` will respond to the overriden
    ``on_move``, etc... methods.
    """

    def __init__(self, track=True):
        """
        :param track: If false, mouseMoveEvents will only be triggered if a
          button is pressed.
        """
        # super().__init__(parent=parent)
        #
        self.setMouseTracking(track)
        self.mousepos_pix = None
        #
        self.last_left_press_pos = None
        self.last_mid_press_pos = None
        self.last_right_press_pos = None
        #
        self.last_left_release_pos = None
        self.last_mid_release_pos = None
        self.last_right_release_pos = None
        #
        self.last_move_pos = None

    def mousePressEvent(self, event):
        """
        Wheel click event handler
        """
        b = event.button()
        if b == QtCore.Qt.MouseButton.LeftButton:
            self.last_left_press_pos = event.pos()
            self.on_left_press(event)
        elif b == QtCore.Qt.MouseButton.MidButton:
            self.last_mid_press_pos = event.pos()
            self.on_mid_press(event)
        elif b == QtCore.Qt.MouseButton.RightButton:
            self.last_right_press_pos = event.pos()
            self.on_right_press(event)

    def mouseReleaseEvent(self, event):
        """
        Wheel release event handler
        """
        b = event.button()
        if b == QtCore.Qt.MouseButton.LeftButton:
            self.last_left_release_pos = event.pos()
            self.on_left_release(event)
        elif b == QtCore.Qt.MouseButton.MidButton:
            self.last_mid_release_pos = event.pos()
            self.on_mid_release(event)
        elif b == QtCore.Qt.MouseButton.RightButton:
            self.last_right_release_pos = event.pos()
            self.on_right_release(event)

    def mouseMoveEvent(self, event):
        """
        """
        p = event.pos()
        # this does happen! filter out
        if p == self.last_move_pos:
            return
        # continue if we had true movement
        bbb = event.buttons()
        # button codes: https://doc.qt.io/qt-5/qt.html#MouseButton-enum
        has_l = bool(bbb & QtCore.Qt.LeftButton)
        has_m = bool(bbb & QtCore.Qt.MidButton)
        has_r = bool(bbb & QtCore.Qt.RightButton)
        #
        self.on_move(event, has_l, has_m, has_r, p, self.last_move_pos)
        self.last_move_pos = p

    def wheelEvent(self, event):
        """
        Override me. This is a simple wrapper, but may include functionality
        like storing positions if needed.
        """
        mods = event.modifiers()
        has_ctrl = bool(mods & QtCore.Qt.ControlModifier)
        has_alt = bool(mods & QtCore.Qt.AltModifier)
        has_shift = bool(mods & QtCore.Qt.ShiftModifier)
        self.on_wheel(event, has_ctrl, has_alt, has_shift)

    def on_left_press(self, event):
        """
        Override me
        """
        pass

    def on_mid_press(self, event):
        """
        Override me
        """
        pass

    def on_right_press(self, event):
        """
        Override me
        """
        pass

    def on_left_release(self, event):
        """
        Override me
        """
        pass

    def on_mid_release(self, event):
        """
        Override me
        """
        pass

    def on_right_release(self, event):
        """
        Override me
        """
        pass

    def on_wheel(self, event, has_ctrl, has_alt, has_shift):
        """
        Override me
        """
        pass

    def on_move(self, event, has_left, has_mid, has_right, this_pos, last_pos):
        """
        Override me
        """
        pass
